/*  TA3D, a remake of Total Annihilation
    Copyright (C) 2005  Roland BROCHARD

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

/*-----------------------------------------------------------------------------------\
|                                         hpi.cpp                                    |
|  ce fichier contient les structures, classes et fonctions nécessaires à la lecture |
| des fichiers hpi/ufo/... de total annihilation mais sans fournir les fonctions de  |
| lecture des fichiers contenus dans ces archives!                                   |
|                                                                                    |
\-----------------------------------------------------------------------------------*/

#include <allegro.h>
#include "console.h"
#include "hpi.h"

HPIHeader read_hpi_header(char *filename)			// Lit l'entête d'un fichier hpi
{
	HPIHeader header;
	FILE *f;
	f=fopen(filename,"rb");
	fread(&header,sizeof(header),1,f);
	fclose(f);
	return header;
}

	int HPI_FILE::Decompress(char *out, char *in, int len)
	{
	/*
		Decomprime le tampon "in" de taille "len" octets dans le tampon "out" (précédemment
		alloué) retourne le nombre d'octets décomprimés.
	*/

		int x;
		int outbufptr;
		int mask;
		int tag;
		int inptr;
		int outptr;
		int count;
		int done;
		char Window[4096];
		int inbufptr;

		done = FALSE;

		inptr = 0;
		outptr = 0;
		outbufptr = 1;
		mask = 1;
		tag = in[inptr++];
	
		while (!done) {
			if ((mask & tag) == 0) {
				out[outptr++] = in[inptr];
				Window[outbufptr] = in[inptr];
				outbufptr = (outbufptr + 1) & 0xFFF;
				inptr++;
				}
			else {
				count = *((unsigned short *) (in+inptr));
				inptr += 2;
				inbufptr = count >> 4;
				if (inbufptr == 0)
					return outptr;
				else {
					count = (count & 0x0f) + 2;
					if (count >= 0) {
						for (x = 0; x < count; x++) {
							out[outptr++] = Window[inbufptr];
							Window[outbufptr] = Window[inbufptr];
							inbufptr = (inbufptr + 1) & 0xFFF;
							outbufptr = (outbufptr + 1) & 0xFFF;
							}
						}
					}
				}
			mask *= 2;
			if (mask & 0x0100) {
				mask = 1;
				tag = in[inptr++];
				}
			}
		return outptr;
	}

	int HPI_FILE::read_and_decrypt(FILE *f,byte *dat,int fpos,int size)
	/*
		lit et décrypte le contenu du fichier f dont l'entête a été lue dans header
	*/
	{
		int Key= ~((header.HeaderKey * 4) | (header.HeaderKey >> 6));
		int count;
		int tkey;
		int result;

		/* commence par positionner le curseur dans le fichier */
		fseek(f, fpos, SEEK_SET);

		/* lit les données dans le tampon */
		result = fread(dat, 1, size, f);

		/* pour chaque caractère du tampon... */
		if(header.HeaderKey)						// Décrypte les données si elles sont cryptées
			for (count = 0; count < size; count++) {

				/* calcule la clef de décryptage tkey = (fpos + count) XOR Key */
				tkey = (fpos + count) ^ Key;

				/* puis décode l'octet:  
				data[count] = tkey XOR (NOT data[count]) */
				dat[count] = tkey ^ ~dat[count];
				}

		/* result est le nombre d'octets lus depuis le fichier, 
		et devrait être égal à header.DirectorySize */
		return result;
	}

char *find_hpi(char *file)			// Cherche l'archive qui contient le fichier file
{
	char *hpi_name=NULL;
	al_ffblk search;

	if(al_findfirst("*.hpi",&search,0)==0)
	do
	{
		HPI_FILE hpi;
		hpi.load(search.name);
		if(hpi.exist(file)) {
			hpi.destroy();
			hpi_name=strdup(search.name);
			al_findclose(&search);
			return hpi_name;
			}
		hpi.destroy();
	}while(al_findnext(&search)==0);
	al_findclose(&search);
	if(al_findfirst("*.ufo",&search,0)==0)
	do
	{
		HPI_FILE hpi;
		hpi.load(search.name);
		if(hpi.exist(file)) {
			hpi.destroy();
			hpi_name=strdup(search.name);
			al_findclose(&search);
			return hpi_name;
			}
		hpi.destroy();
	}while(al_findnext(&search)==0);
	al_findclose(&search);
	if(al_findfirst("*.ccx",&search,0)==0)
	do
	{
		HPI_FILE hpi;
		hpi.load(search.name);
		if(hpi.exist(file)) {
			hpi.destroy();
			hpi_name=strdup(search.name);
			al_findclose(&search);
			return hpi_name;
			}
		hpi.destroy();
	}while(al_findnext(&search)==0);
	al_findclose(&search);
	if(al_findfirst("*.gp3",&search,0)==0)
	do
	{
		HPI_FILE hpi;
		hpi.load(search.name);
		if(hpi.exist(file)) {
			hpi.destroy();
			hpi_name=strdup(search.name);
			al_findclose(&search);
			return hpi_name;
			}
		hpi.destroy();
	}while(al_findnext(&search)==0);
	al_findclose(&search);
	return hpi_name;
}

byte *load_file(char *filename)			// Trouve l'archive contenant le fichier demandé et en extrait le fichier cherché
{
	HPI_FILE hpi;

	al_ffblk search;

	if(al_findfirst("*.hpi",&search,0)==0)
	do
	{
		hpi.load(search.name);
		if(hpi.exist(filename)) {		// Si on a trouvé le fichier
			byte *data=hpi.extract_memory_file(filename);
			hpi.destroy();
			al_findclose(&search);
			return data;
			}
		hpi.destroy();
	}while(al_findnext(&search)==0);
	al_findclose(&search);
	if(al_findfirst("*.ufo",&search,0)==0)
	do
	{
		hpi.load(search.name);
		if(hpi.exist(filename)) {		// Si on a trouvé le fichier
			byte *data=hpi.extract_memory_file(filename);
			hpi.destroy();
			al_findclose(&search);
			return data;
			}
		hpi.destroy();
	}while(al_findnext(&search)==0);
	al_findclose(&search);
	if(al_findfirst("*.ccx",&search,0)==0)
	do
	{
		hpi.load(search.name);
		if(hpi.exist(filename)) {		// Si on a trouvé le fichier
			byte *data=hpi.extract_memory_file(filename);
			hpi.destroy();
			al_findclose(&search);
			return data;
			}
		hpi.destroy();
	}while(al_findnext(&search)==0);
	al_findclose(&search);

	hpi.load("rev31.gp3");
	if(hpi.exist(filename)) {		// Si on a trouvé le fichier
		byte *data=hpi.extract_memory_file(filename);
		hpi.destroy();
		return data;
		}
	hpi.destroy();
	return NULL;
}

int get_nb_file(HPI_FILE *hpi,char *dir,char *ext)
{
	int n=0,i=0;
	char *name=NULL;
	while(name=hpi->find(dir,i++)) {
		if(strstr(name,ext))
			n++;
		delete name;
		}
	return n;
}

char **file_list(char *dir,char *ext,int *nb)
{
	int n=0;
	HPI_FILE hpi;

	al_ffblk search;

	if(al_findfirst("*.hpi",&search,0)==0)
	do
	{
		hpi.load(search.name);
		n+=get_nb_file(&hpi,dir,ext);
		hpi.destroy();
	}while(al_findnext(&search)==0);
	al_findclose(&search);
	if(al_findfirst("*.ufo",&search,0)==0)
	do
	{
		hpi.load(search.name);
		n+=get_nb_file(&hpi,dir,ext);
		hpi.destroy();
	}while(al_findnext(&search)==0);
	al_findclose(&search);
	if(al_findfirst("*.ccx",&search,0)==0)
	do
	{
		hpi.load(search.name);
		n+=get_nb_file(&hpi,dir,ext);
		hpi.destroy();
	}while(al_findnext(&search)==0);
	al_findclose(&search);

	hpi.load("rev31.gp3");
	n+=get_nb_file(&hpi,dir,ext);
	hpi.destroy();

	char **list=(char**) malloc(sizeof(char*)*n);
	char *name=NULL;
	int i;
	int p=0;
	if(al_findfirst("*.hpi",&search,0)==0)
	do
	{
		hpi.load(search.name);
		i=0;
		while(name=hpi.find(dir,i++)) {
			if(strstr(name,ext))
				list[p++]=name;
			else
				delete name;
			}
		hpi.destroy();
	}while(al_findnext(&search)==0);
	al_findclose(&search);
	if(al_findfirst("*.ufo",&search,0)==0)
	do
	{
		hpi.load(search.name);
		i=0;
		while(name=hpi.find(dir,i++)) {
			if(strstr(name,ext))
				list[p++]=name;
			else
				delete name;
			}
		hpi.destroy();
	}while(al_findnext(&search)==0);
	al_findclose(&search);
	if(al_findfirst("*.ccx",&search,0)==0)
	do
	{
		hpi.load(search.name);
		i=0;
		while(name=hpi.find(dir,i++)) {
			if(strstr(name,ext))
				list[p++]=name;
			else
				delete name;
			}
		hpi.destroy();
	}while(al_findnext(&search)==0);
	al_findclose(&search);

	hpi.load("rev31.gp3");
		i=0;
		while(name=hpi.find(dir,i++)) {
			if(strstr(name,ext))
				list[p++]=name;
			else
				delete name;
			}
	hpi.destroy();
	*nb=n;
	return list;
}
