/*  TA3D, a remake of Total Annihilation
    Copyright (C) 2005  Roland BROCHARD

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

#include <allegro.h>		// Allegro
#include <alleggl.h>		// AllegroGL
#include <GL/glu.h>			// GLU

#include "intro.h"			// Introduction

#define DEBUG_MODE				// Mode débogage

#include "console.h"		// Console
#include "ta3d.h"			// Moteur
#include "menu.h"			// Menus du jeu

FONT *aglfont;
RGB *pal;

VECTOR cursor_on_map(CAMERA *cam,MAP *map);

int main(int argc,char *argv[])
{
allegro_init();

install_timer();			// Initialise allegro
install_keyboard();
install_mouse();
install_allegro_gl();

install_int_ex(Timer,precision);

	allegro_gl_clear_settings();			// Initialise AllegroGL
	allegro_gl_set (AGL_SAMPLE_BUFFERS, 0);
	allegro_gl_set (AGL_SAMPLES, 0);
	allegro_gl_set (AGL_COLOR_DEPTH, 32);
	allegro_gl_set (AGL_Z_DEPTH, 32);
	allegro_gl_set (AGL_FULLSCREEN, TRUE);
	allegro_gl_set (AGL_DOUBLEBUFFER, 1);
	allegro_gl_set (AGL_RENDERMETHOD, 1);
	allegro_gl_set (AGL_SUGGEST, AGL_RENDERMETHOD | AGL_COLOR_DEPTH | AGL_Z_DEPTH | AGL_DOUBLEBUFFER | AGL_FULLSCREEN | AGL_SAMPLES | AGL_SAMPLE_BUFFERS);

	allegro_gl_use_mipmapping(TRUE);

	allegro_gl_flip_texture(false);

request_refresh_rate(70);

set_gfx_mode(GFX_OPENGL_WINDOWED, 640, 480, 0, 0);		// Entre en mode graphique OpenGL

install_ext();		// Pour les fonctions OpenGl supplémentaires

if(g_useTextureCompression)			// Active la compression de texture
	allegro_gl_set_texture_format(GL_COMPRESSED_RGB_ARB);
else
	allegro_gl_set_texture_format(GL_RGB8);

set_uformat(U_ASCII);		// Juste histoire d'avoir un affichage correct des textes

set_window_title("Total Annihilation 3D");

aglfont=allegro_gl_convert_allegro_font(font,AGL_FONT_TYPE_TEXTURED,-0.9f);

start=Atimer;		// Pour l'animation du curseur

Console=new CONSOLE;				// Initialise la console

Console->StartRecording("console.log");
Console->AddEntry("TA3D Engine");
Console->AddEntry(TA3D_ENGINE_VERSION);
print_system_data(Console);

Console->AddEntry("Lancement de l'introduction");

play_intro();

Console->AddEntry("Lancement d'une partie");
model_manager.init();
unit_manager.init();
feature_manager.init();

pal=LoadPal("/palettes/palette.pal");			// Charge la palette graphique
set_palette(pal);		// Active la palette chargée
{
	byte *data=load_file("/anims/CURSORS.GAF");
	cursor.load_gaf(data);
	cursor.convert();
	free(data);
}

if(argc==1)
	main_menu();
else {
	for(int i=0;i<strlen(argv[1]);i++)
		if(argv[1][i]=='_')
			argv[1][i]=' ';
	char *hpi_name=find_hpi(argv[1]);
	if(hpi_name)						// Charge l'archive contenant la carte
		play(argv[1]);
	}

cursor.destroy();

Console->AddEntry("Fin du programme");

Console->destroy();		// Détruit la console
delete Console;

allegro_exit();

return 0;
}
END_OF_MAIN();

/*--------------------------------------------------------------------\
|                            Moteur du jeu                            |
\--------------------------------------------------------------------*/
void play(char *filename)
{
al_ffblk search;

loading(0.0f,"Chargement en cours des textures");

/*-----------------------charge les textures-------------------------*/

if(al_findfirst("*.hpi",&search,0)==0)
do
{
	texture_manager.all_texture_in_hpi(search.name);
}while(al_findnext(&search)==0);
al_findclose(&search);
loading(100.0f/21.0f,"Chargement en cours des textures");
if(al_findfirst("*.ufo",&search,0)==0)
do
{
	texture_manager.all_texture_in_hpi(search.name);
}while(al_findnext(&search)==0);
al_findclose(&search);

loading(200.0f/21.0f,"Chargement en cours des textures");

texture_manager.all_texture_in_hpi("rev31.gp3");

loading(100.0f/7.0f,"Chargement en cours des modeles 3D");
/*-----------------------charge les modèles 3D-----------------------*/

model_manager.init();
if(al_findfirst("*.hpi",&search,0)==0)
do
{
	model_manager.load_all_in_hpi(search.name);
}while(al_findnext(&search)==0);
al_findclose(&search);
loading(100.0f/7.0f+100.0f/21.0f,"Chargement en cours des modeles 3D");
if(al_findfirst("*.ufo",&search,0)==0)
do
{
	model_manager.load_all_in_hpi(search.name);
}while(al_findnext(&search)==0);
al_findclose(&search);

loading(100.0f/7.0f+200.0f/21.0f,"Chargement en cours des modeles 3D");

model_manager.load_all_in_hpi("rev31.gp3");

loading(200.0f/7.0f,"Chargement en cours des elements graphiques");
/*-----------------------charge les éléments graphiques--------------*/

load_features();
feature_manager.clean();

loading(300.0f/7.0f,"Chargement en cours des unites");
/*-----------------------charge les unités---------------------------*/

unit_manager.init();
int nb_inconnu=0;
if(al_findfirst("*.hpi",&search,0)==0)
do
{
	nb_inconnu+=unit_manager.all_units_in_hpi(search.name);
}while(al_findnext(&search)==0);
al_findclose(&search);
loading(300.0f/7.0f+100.0f/21.0f,"Chargement en cours des unites");
if(al_findfirst("*.ufo",&search,0)==0)
do
{
	nb_inconnu+=unit_manager.all_units_in_hpi(search.name);
}while(al_findnext(&search)==0);
al_findclose(&search);

loading(300.0f/7.0f+200.0f/21.0f,"Chargement en cours des unites");

nb_inconnu+=unit_manager.all_units_in_hpi("rev31.gp3");

printf("%d mots clées inconnus\n",nb_inconnu);

unit_manager.Identify();

loading(400.0f/7.0f,"Libere la memoire inutilisee");
/*-----------------------libère la mémoire inutilisée----------------*/

texture_manager.destroy();

loading(500.0f/7.0f,"Initialise le moteur");
/*-----------------------initialise le moteur------------------------*/

SetDefOpenGlState();
particle_engine.init();

set_palette(pal);

PLAYERS	players;		// Objet contenant les données sur les joueurs
players.add("joueur0","arm");		// Ajoute un joueur

units.init();
units.create(unit_manager.get_unit_index("ARMCOM"),0);

loading(600.0f/7.0f,"Charge la carte");
/*-----------------------charge la carte-----------------------------*/

byte *map_file=load_file(filename);

if(!map_file)	return;
MAP *map=load_tnt_map(map_file);
free(map_file);

GLuint	sky=LoadTex("sky.tga");
GLuint	glow=LoadTex("glow.tga");

loading(100.0f,"Chargement termine");

float Conv=1.0f/SECS_TO_TIMER(1);
float dt=0.0f;
float t=0.0f;
int count=Atimer;
int nbfps=0;
int fpscount=Atimer;
int fps=0;
int i;

float r1,r2,r3;
r1=r2=r3=0.0f;
r1=-45.0f;
CAMERA cam;

cam.RPos.y=150.0f;
cam.zfar=500.0f;

float FogD=0.3f;

float FogColor[]={0.8f,0.8f,0.8f,1.0f};
	glFogi (GL_FOG_MODE, GL_LINEAR);
	glFogfv (GL_FOG_COLOR, FogColor);
	glFogf (GL_FOG_DENSITY, FogD);
	glHint (GL_FOG_HINT, GL_NICEST);
	glFogf (GL_FOG_START, cam.zfar*0.5f);
	glFogf (GL_FOG_END, cam.zfar);
	
glEnable(GL_FOG);

HWLIGHT	sun;
sun.Att=0.0f;
sun.Dir.x=-1.0f;
sun.Dir.y=1.0f;
sun.Dir.z=1.0f;
sun.Dir.Unit();
sun.LightAmbient[0]=0.25f;
sun.LightAmbient[1]=0.25f;
sun.LightAmbient[2]=0.25f;
sun.LightAmbient[3]=0.25f;
sun.LightDiffuse[0]=1.0f;
sun.LightDiffuse[1]=1.0f;
sun.LightDiffuse[2]=1.0f;
sun.LightDiffuse[3]=1.0f;
sun.LightSpecular[0]=0.0f;
sun.LightSpecular[1]=0.0f;
sun.LightSpecular[2]=0.0f;
sun.LightSpecular[3]=0.0f;
sun.Directionnal=true;

float speed_limit=-1.0f;
float delay=1.0f/speed_limit;
int nb_shoot=0;
bool shoot=false;

bool tilde=false;
bool showfps=false;
bool done=false;

fire=particle_engine.addtex("fire.tga","firea.tga");
build_part=particle_engine.addtex("part.tga",NULL);

int mx,my,omb=mouse_b,omb2=mouse_b,omb3=mouse_b;
int	sel_x[2],sel_y[2],selecting=false;
int cur_sel=-1;
bool selected=false;
int	build=-1;				// Indique si l'utilisateur veut construire quelque chose
int cur_sel_index=-1;
int omz=mouse_z;
float cam_h=100.0f;
float time_factor=1.0f;

bool wireframe=false;		// Mode fil de fer pour le terrain
bool shadow=false;			// Active les ombres
bool particle=true;
bool trees=true;

do
{
	r1+=(mouse_z-omz)*0.5f;
	if(r1>-45.0f) r1=-45.0f;
	else if(r1<-75.0f) r1=-75.0f;
	else
		cam_h-=(mouse_z-omz);
	omz=mouse_z;

	cursor_type=CURSOR_DEFAULT;			// Revient au curseur par défaut

	if(mouse_x>128 && mouse_y>32 && mouse_y<SCREEN_H-32) {
		if(selected)
			for(i=0;i<=units.last_index;i++)
				if(units.unit[i].flags!=0 && units.unit[i].owner_id==0 && units.unit[i].sel && unit_manager.unit_type[units.unit[i].type_id].canmove) {
					cursor_type=CURSOR_MOVE;
					break;
					}
		}

	if(build>=0)
		cursor_type=CURSOR_DEFAULT;

	do
	{
		dt=(Atimer-count)*Conv;
		rest(0);
	}while(dt<delay);

	t=Atimer*Conv;
	nbfps++;
	if(nbfps>=10 && (Atimer-fpscount)*Conv>=0.05f) {
		fps=(int)(nbfps/((Atimer-fpscount)*Conv));
		fpscount=Atimer;
		nbfps=0;
		}
	count=Atimer;
	SetDefOpenGlState();
	glClearColor(FogColor[0],FogColor[1],FogColor[2],FogColor[3]);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);		// Efface l'écran

	glFogi (GL_FOG_MODE, GL_LINEAR);
	glFogfv (GL_FOG_COLOR, FogColor);
	glFogf (GL_FOG_DENSITY, FogD);
	glHint (GL_FOG_HINT, GL_NICEST);
	glFogf (GL_FOG_START, cam.zfar*0.5f);
	glFogf (GL_FOG_END, cam.zfar);

/*------------bloc regroupant ce qui est relatif aux commandes----------------*/

	if(mouse_x<128.0f && mouse_y<128.0f && mouse_x>=0.0f && mouse_y>=0.0f && mouse_b==1) {
		cam.RPos.x=(mouse_x-64)*map->map_w/128.0f*252.0f/map->mini_w;
		cam.RPos.z=(mouse_y-64)*map->map_h/128.0f*252.0f/map->mini_h;
		}
	if(mouse_x<10)
		cam.RPos.x-=100.0f*dt;
	if(mouse_x>SCREEN_W-10)
		cam.RPos.x+=100.0f*dt;
	if(mouse_y<10)
		cam.RPos.z-=100.0f*dt;
	if(mouse_y>SCREEN_H-10)
		cam.RPos.z+=100.0f*dt;

	if(mouse_b==4) {
		get_mouse_mickeys(&mx,&my);
		if(omb==mouse_b) {
			cam.RPos.x+=mx;
			cam.RPos.z+=my;
			}
		position_mouse(SCREEN_W>>1,SCREEN_H>>1);
		}
	else if(omb==4)
		position_mouse(SCREEN_W>>1,SCREEN_H>>1);
	omb=mouse_b;

	if(key[KEY_UP])
		cam.RPos.z-=100.0f*dt;
	if(key[KEY_DOWN])
		cam.RPos.z+=100.0f*dt;
	if(key[KEY_RIGHT])
		cam.RPos.x+=100.0f*dt;
	if(key[KEY_LEFT])
		cam.RPos.x-=100.0f*dt;

	if(cam.RPos.x<-0.5f*map->map_w)	cam.RPos.x=-0.5f*map->map_w;
	if(cam.RPos.x>0.5f*map->map_w)	cam.RPos.x=0.5f*map->map_w;
	if(cam.RPos.z<-0.5f*map->map_h+200.0f)	cam.RPos.z=-0.5f*map->map_h+200.0f;
	if(cam.RPos.z>0.5f*map->map_h)	cam.RPos.z=0.5f*map->map_h;

	{
		float h=map->get_unit_h(cam.RPos.x,cam.RPos.z);
		if(h<map->sealvl)
			h=map->sealvl;
		cam.RPos.y+=(h+cam_h-cam.RPos.y)*dt;
	}

	MATRIX_4x4 Rotation;
	r1+=cam.Rotat.x*dt;
	r2+=cam.Rotat.y*dt;
	r3+=cam.Rotat.z*dt;
	Rotation=RotateX(r1*PI/180.0f)*RotateY(r2*PI/180.0f)*RotateZ(r3*PI/180.0f);

	cam.SetMatrix(Rotation,dt);

	if(selected) {
		bool builders=false;
		for(i=0;i<=units.last_index;i++)
			if(units.unit[i].flags!=0 && units.unit[i].owner_id==0 && units.unit[i].sel) {
				builders|=unit_manager.unit_type[units.unit[i].type_id].Builder;
				}
		int pointing=units.pick(&cam);		// Sur quoi le curseur est-il pointé??

		if(pointing>=0) {	// S'il y a quelque chose sous le curseur
			cursor_type=CURSOR_CROSS;
			if(units.unit[pointing].owner_id!=0)	cursor_type=CURSOR_ATTACK;
			else if(units.unit[pointing].port[BUILD_PERCENT_LEFT] && builders)	cursor_type=CURSOR_REPAIR;

			if(mouse_b==1 && omb3!=1)
				switch(cursor_type)
				{
				case CURSOR_ATTACK:		break;
				case CURSOR_REPAIR:
					for(i=0;i<=units.last_index;i++)
						if(units.unit[i].flags!=0 && units.unit[i].owner_id==0 && units.unit[i].sel) {
							if(key[KEY_LSHIFT])
								units.unit[i].add_mission(MISSION_REPAIR,&(units.unit[pointing].Pos),false,0,&(units.unit[pointing]));
							else
								units.unit[i].set_mission(MISSION_REPAIR,&(units.unit[pointing].Pos),false,0,true,&(units.unit[pointing]));
							}
					break;
				};
			}
		}

	if(cursor_type==CURSOR_MOVE && mouse_b==1 && omb3 !=1 && mouse_x>128.0f && mouse_y>32.0f && mouse_y<SCREEN_H-32) {	// Le curseur donne un ordre
		units.give_order_move(0,cursor_on_map(&cam,map),!key[KEY_LSHIFT]);
		}

	if(build>=0 && cursor_type==CURSOR_DEFAULT && mouse_b==1 && omb3 !=1 && mouse_x>128.0f && mouse_y>32.0f && mouse_y<SCREEN_H-32) {	// Le curseur donne un ordre
		units.give_order_build(0,build,cursor_on_map(&cam,map),!key[KEY_LSHIFT]);
		if(!key[KEY_LSHIFT])
			build=-1;
		}

//---------------------------------	Code de sélection d'unités

	if(mouse_x>128 && mouse_y>32 && mouse_y<SCREEN_H-32) {
		if(mouse_b==2 && omb3!=2) {					// Déselectionne les unités
			if(build>=0)
				build=-1;				// Sort du mode construction
			else {
				selected=false;
				cur_sel=-1;
				cur_sel_index=-1;
				for(i=0;i<=units.last_index;i++)
					if(units.unit[i].owner_id==0)			// On peut désélectionner les morts, ça ne change rien :-)
						units.unit[i].sel=false;
				}
			}
		}

	if(build==-1 && (cursor_type==CURSOR_DEFAULT || cursor_type==CURSOR_CROSS) && mouse_x>128 && (mouse_y>32 || selecting)
	   && (mouse_y<SCREEN_H-32 || selecting)) {		// Si le curseur est dans la zone de jeu
		if(mouse_b!=1 && selecting) {		// Récupère les unités présentent dans la sélection
			if(sel_x[0]==sel_x[1] && sel_y[0]==sel_y[1]) {
				int pointing=units.pick(&cam);		// Sélectionne une unité sur clic
				if(!key[KEY_LSHIFT])
					for(i=0;i<=units.last_index;i++)
						if(units.unit[i].flags!=0 && units.unit[i].owner_id==0)
							units.unit[i].sel=false;
				if(pointing>=0 && units.unit[pointing].port[BUILD_PERCENT_LEFT]==0.0f)		// On ne sélectionne pas les unités en construction
					units.unit[pointing].sel^=true;			// Sélectionne/Désélectionne si l'unité est déjà sélectionnée en appuyant sur SHIFT
				selected=false;
				for(i=0;i<=units.last_index;i++)
					if(units.unit[i].flags!=0 && units.unit[i].owner_id==0)
						selected|=units.unit[i].sel;
				}
			else
				selected=units.select(&cam,sel_x,sel_y);		// Séléction au lasso
			cur_sel=-1;
			cur_sel_index=-1;
			for(i=0;i<=units.last_index && cur_sel!=-2;i++)
				if(units.unit[i].flags!=0 && units.unit[i].owner_id==0 && units.unit[i].sel)
					cur_sel= (cur_sel==-1) ? i : -2;
			if(cur_sel>=0) {
				cur_sel_index=cur_sel;
				cur_sel=units.unit[cur_sel].type_id;
				}
			}
		selecting=false;
		if(mouse_b==1) {
			if(omb3!=1) {
				sel_x[0]=mouse_x;
				sel_y[0]=mouse_y;
				}
			sel_x[1]=mouse_x;
			sel_y[1]=mouse_y;
			selecting=true;
			}
		}
	else
		selecting=false;
	omb3=mouse_b;

	if(mouse_x<=128 || mouse_y<=32 || mouse_y>=SCREEN_H-32)	cursor_type=CURSOR_DEFAULT;

/*--------------bloc regroupant ce qui est relatif au temps-------------------*/

	float timetosimulate=dt*time_factor;								// Moteur physique
	for(float dt_sim=0.0f;dt_sim<timetosimulate;dt_sim+=TIME_UNIT) {
		float dt_r=min(timetosimulate-dt_sim,TIME_UNIT);
		if(particle)
			particle_engine.move(dt_r);
		units.move(dt_r,map);
		if(trees)
			features.move(dt_r);
		}

/*----------------------------------------------------------------------------*/

	cam.SetView();

	sun.Set(cam);
	sun.Enable();

	cam.SetView();
	glRotatef(Atimer*Conv*180.0f/PI,0.0f,1.0f,0.0f);
	glTranslatef(0.0f,60.0f,0.0f);

	cam.SetView();
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D,sky);
	glDisable(GL_CULL_FACE);
	glDisable(GL_LIGHTING);
	glDisable(GL_FOG);
	glBegin(GL_QUADS);
		glTexCoord2f(0.0f,0.0f);		glVertex3f(-2.0f*map->map_w,300.0f,-2.0f*map->map_h);
		glTexCoord2f(1.0f,0.0f);		glVertex3f(2.0f*map->map_w,300.0f,-2.0f*map->map_h);
		glTexCoord2f(1.0f,1.0f);		glVertex3f(2.0f*map->map_w,300.0f,2.0f*map->map_h);
		glTexCoord2f(0.0f,1.0f);		glVertex3f(-2.0f*map->map_w,300.0f,2.0f*map->map_h);
	glEnd();
	glEnable(GL_CULL_FACE);
	glEnable(GL_LIGHTING);
	glEnable(GL_FOG);

	units.draw(&cam);

	if(build>=0 && mouse_x>128 && mouse_y>32 && mouse_y<SCREEN_H-32) {			// Affiche le bâtiment à construire
		VECTOR target=cursor_on_map(&cam,map);
		cam.SetView();
		glTranslatef(target.x,target.y,target.z);
		glScalef(unit_manager.unit_type[build].Scale,unit_manager.unit_type[build].Scale,unit_manager.unit_type[build].Scale);
		unit_manager.unit_type[build].model->draw();
		}

	if(wireframe)
		glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);

	map->draw(&cam,false,0.0f,t,dt*time_factor);

	if(wireframe)
		glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);

	if(selected && key[KEY_LSHIFT])
		for(i=0;i<=units.last_index;i++)
			if(units.unit[i].flags!=0 && units.unit[i].owner_id==0 && units.unit[i].sel)
				units.unit[i].show_orders();					// Dessine les ordres reçus par l'unité

	if(trees)
		features.draw(&cam);		// Dessine les éléments "2D"

	if(particle)
		particle_engine.draw(&cam,map->map_w,map->map_h,map->bloc_w,map->bloc_h,map->view);	// Dessine les particules

	if(shadow)
		units.draw_shadow(&cam,-1000.0f*sun.Dir);

	if(key[KEY_ESC]) done=true;
	if(key[KEY_TILDE]) {
		if(!tilde)
			Console->show^=true;
		tilde=true;
		}
	else
		tilde=false;

	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	for(i=7;i>=0;i--) {
		glActiveTextureARB(GL_TEXTURE0_ARB+i);
		ReInitTexSys();
		glDisable(GL_TEXTURE_2D);
		}
	allegro_gl_set_allegro_mode();		// Affiche console, infos,...

/*---------------------code expérimental---------------------------------*/
/*	ReInitTexSys();
	glBlendFunc(GL_SRC_ALPHA,GL_ONE);
	glEnable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D,glow);
	for(int y=0;y<SCREEN_H;y+=16)
		for(int x=0;x<SCREEN_W;x+=16) {
			int c=getpixel(screen,x,y);
			int r=getr(c);
			int g=getg(c);
			int b=getb(c);
			if(r>4 && g<(r>>2) && b<(r>>2)) {
				glColor4f(r/255.0f,g/255.0f,b/255.0f,0.5f);
				glBegin(GL_QUADS);
					glTexCoord2f(0.0f,0.0f);	glVertex2i(x-16.0f,SCREEN_H-y-16.0f);
					glTexCoord2f(1.0f,0.0f);	glVertex2i(x+16.0f,SCREEN_H-y-16.0f);
					glTexCoord2f(1.0f,1.0f);	glVertex2i(x+16.0f,SCREEN_H-y+16.0f);
					glTexCoord2f(0.0f,1.0f);	glVertex2i(x-16.0f,SCREEN_H-y+16.0f);
				glEnd();
				}
			}
	glColor4f(1.0f,1.0f,1.0f,1.0f);
	glBlendFunc(GL_ONE,GL_ONE);
	glDisable(GL_BLEND);
	allegro_gl_set_allegro_mode();		// Affiche console, infos,...*/
/*-----------------------------------------------------------------------*/

	if(selecting) {						// Affiche le rectangle de selection
		glDisable(GL_TEXTURE_2D);
		glBegin(GL_LINE_LOOP);
			glColor4f(0.0f,0.0f,0.0f,1.0f);
			glVertex2i(sel_x[0]+1,sel_y[0]+1);
			glVertex2i(sel_x[1]+1,sel_y[0]+1);
			glVertex2i(sel_x[1]+1,sel_y[1]+1);
			glVertex2i(sel_x[0]+1,sel_y[1]+1);
		glEnd();
		glBegin(GL_LINE_LOOP);
			glColor4f(1.0f,1.0f,1.0f,1.0f);
			glVertex2i(sel_x[0],sel_y[0]);
			glVertex2i(sel_x[1],sel_y[0]);
			glVertex2i(sel_x[1],sel_y[1]);
			glVertex2i(sel_x[0],sel_y[1]);
		glEnd();
		}

	int n=cur_sel;
	if(n>=0 && units.unit[cur_sel_index].port[BUILD_PERCENT_LEFT]>0.0f)		// Unité non terminée
		n=-1;
	int sel=unit_manager.unit_build_menu(n,omb2);		// Menu correspondant à l'unité
	units.complete_menu(cur_sel_index);
	units.menu_order(omb2);
	players.show_resources();

	if(sel>=0 && mouse_b==2 && omb2!=2) {
		MISSION *cur=units.unit[cur_sel_index].mission;
		MISSION **old=&(units.unit[cur_sel_index].mission);
		int nb=1;
		if(key[KEY_LSHIFT]) nb=5;
		if(cur)
			while(cur->next) {
				if(cur->mission==MISSION_BUILD && cur->data==sel) {			// Efface un ordre
					*old=cur->next;
					free(cur);
					nb--;
					if(nb==0)
						break;
					}
				old=&(cur->next);
				cur=cur->next;
				}
		}
	if(sel>=0 && mouse_b==1 && omb2!=1) {
		build=sel;
		if(unit_manager.unit_type[cur_sel].TEDclass==CLASS_PLANT) {			// S'il s'agit d'un bâtiment
			if(key[KEY_LSHIFT])
				for(i=0;i<5;i++)
					units.give_order_build(0,build,units.unit[cur_sel_index].Pos,false);
			else
				units.give_order_build(0,build,units.unit[cur_sel_index].Pos,false);
			build=-1;
			}
		}

	omb2=mouse_b;

	map->draw_mini(0,0,128,128,&cam);	// Mini-carte
	units.draw_mini(map->map_w,map->map_h,map->mini_w,map->mini_h);

	glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);
	glEnable(GL_BLEND);
	if(cur_sel>=0 && unit_manager.unit_type[cur_sel].script) {
		float Y=32.0f;
		allegro_gl_printf(aglfont,128.0f,Y,0.0f,0xFFFFFF,"%d scripts",unit_manager.unit_type[cur_sel].script->nb_script);	Y+=9.0f;
		for(i=0;i<unit_manager.unit_type[cur_sel].script->nb_script;i++) {
			allegro_gl_printf(aglfont,128.0f,Y,0.0f,0xFFFFFF,"%d %s",i,unit_manager.unit_type[cur_sel].script->name[i]);	Y+=9.0f;
			}
		}

	glDisable(GL_BLEND);

	char *cmd=NULL;
	if(!shoot)
		cmd=Console->draw(aglfont,dt,text_height(font));			// Affiche la console

	float Y=0.0f;

	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);

	if(showfps) {
		allegro_gl_printf(aglfont,0.0f,Y,0.0f,0xFFFFFF,"fps: %d",fps);
		Y+=9.0f;
		}

	glDisable(GL_BLEND);

	if(mouse_b!=4)
		draw_cursor();

	allegro_gl_unset_allegro_mode();

	allegro_gl_flip();

	if(shoot) {
		BITMAP *shoot_bmp=create_bitmap(SCREEN_W,SCREEN_H);
		blit(screen,shoot_bmp,0,0,0,0,SCREEN_W,SCREEN_H);
		char nom[100];
		nom[0]=0;
		strcat(nom,"shoot00");
		nom[strlen(nom)-2]+=nb_shoot/10;
		nom[strlen(nom)-1]+=nb_shoot%10;
		nb_shoot=++nb_shoot%100;
		strcat(nom,".pcx");
		save_bitmap(nom,shoot_bmp,NULL);
		destroy_bitmap(shoot_bmp);
		shoot=false;
		}

	if(cmd) {		// Analyse les commandes tapées dans la console
		if(strstr(cmd,"fps_on")) showfps=true;				// Affiche le nombre d'images/seconde
		if(strstr(cmd,"fps_off")) showfps=false;				// Cache le nombre d'images/seconde
		if(strstr(cmd,"shoot")) shoot=true;					// Prend une capture d'écran
		if(strstr(cmd,"exit")) done=true;					// Quitte le programme
		if(strstr(cmd,"wireframe")) wireframe^=true;
		if(strstr(cmd,"shadow")) shadow^=true;
		if(strstr(cmd,"particle")) particle^=true;
		if(strstr(cmd,"trees")) trees^=true;
		if(strstr(cmd,"timefactor "))
			time_factor=atof(strstr(cmd,"timefactor ")+11);
		if(strstr(cmd,"addhp "))
			if(selected) {					// Sur les unités sélectionnées
				int value=atoi(strstr(cmd,"addhp ")+6);
				for(i=0;i<=units.last_index;i++)
					if(units.unit[i].flags!=0 && units.unit[i].owner_id==0 && units.unit[i].sel) {
						units.unit[i].hp+=value;
						if(units.unit[i].hp<0)
							units.unit[i].hp=0;
						else if(units.unit[i].hp>unit_manager.unit_type[units.unit[i].type_id].MaxDamage)
							units.unit[i].hp=unit_manager.unit_type[units.unit[i].type_id].MaxDamage;
						}
				}
		if(strstr(cmd,"reset_script"))							// Réinitialise les scripts
			if(selected) {					// Sur les unités sélectionnées
				for(i=0;i<=units.last_index;i++)
					if(units.unit[i].flags!=0 && units.unit[i].owner_id==0 && units.unit[i].sel)
						units.unit[i].reset_script();
				}
		if(strstr(cmd,"kill"))							// Détruit les unités sélectionnées
			if(selected) {					// Sur les unités sélectionnées
				for(i=0;i<=units.last_index;i++)
					if(units.unit[i].flags!=0 && units.unit[i].owner_id==0 && units.unit[i].sel)
						units.kill(i);
				selected=false;
				cur_sel=-1;
				}
		if(strstr(cmd,"script "))							// Force l'éxecution d'un script
			if(selected) {					// Sur les unités sélectionnées
				int id=atoi(strstr(cmd,"script ")+7);
				for(i=0;i<=units.last_index;i++)
					if(units.unit[i].flags!=0 && units.unit[i].owner_id==0 && units.unit[i].sel)
						units.unit[i].launch_script(id);
				}
		}
	if(key[KEY_F12])
		shoot=true;

}while(!done);

#ifdef RECORD_VIDEO
MyVID.CloseVID();
MyVID.destroy();
#endif

glDeleteTextures(1,&sky);
glDeleteTextures(1,&glow);

printf("nombre de models chargés: %d \n",model_manager.nb_models);
printf("nombre d'unités chargées: %d \n",unit_manager.nb_unit);
model_manager.destroy();
unit_manager.destroy();
particle_engine.destroy();
units.destroy();

map->destroy();
delete(map);

printf("nombre de textures chargées: %d \n",texture_manager.nbtex);

feature_manager.destroy();

features.destroy();

texture_manager.destroy();
}

int anim_cursor(int type)
{
	if(type==-1)
		return (Atimer-start>>17)%cursor.anm[cursor_type].nb_bmp;
	else
		return (Atimer-start>>17)%cursor.anm[type].nb_bmp;
}

void draw_cursor()
{
	int curseur=anim_cursor();
	int dx=cursor.anm[cursor_type].ofs_x[curseur];
	int dy=cursor.anm[cursor_type].ofs_y[curseur];
	int sx=cursor.anm[cursor_type].bmp[curseur]->w-1;
	int sy=cursor.anm[cursor_type].bmp[curseur]->h-1;
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	PutTex(cursor.anm[cursor_type].glbmp[curseur],mouse_x-dx,mouse_y-dy,mouse_x-dx+sx,mouse_y-dy+sy);
	glDisable(GL_BLEND);
}

VECTOR cursor_on_map(CAMERA *cam,MAP *map)
{
	VECTOR cur_dir;
	cur_dir=cam->Dir+2.0f*(mouse_x-SCREEN_W*0.5f)/SCREEN_W*cam->Side-1.5f*(mouse_y-SCREEN_H*0.5f)/SCREEN_H*cam->Up;
	cur_dir.Unit();		// Direction pointée par le curseur
	return map->hit(cam->Pos,cur_dir);
}
